"""
Composable functions for extracting data using 
`lxml <http://lxml.de/>`_.
"""

from __future__ import absolute_import, unicode_literals, print_function
import logging
import wex.py2compat ; assert wex.py2compat
from six import string_types
from six.moves.urllib_parse import urljoin, quote, unquote
from functools import partial
from operator import is_, methodcaller
import codecs

from six.moves import reduce

from lxml.etree import XPath, _ElementTree, Element
from lxml.cssselect import CSSSelector
from lxml.html import XHTML_NAMESPACE, HTMLParser

from .composed import composable
from .cache import cached
from .iterable import flatten, map_if_iter, filter_if_iter
from .ncr import replace_invalid_ncr

SKIP = object()
skip = partial(is_, SKIP)


UNPARSEABLE = Element('unparseable')

base_href = XPath('//base[@href]/@href | //x:base[@href]/@href',
                  namespaces={'x': XHTML_NAMESPACE})

space_join = composable(' '.join)


default_namespaces = {'re': 'http://exslt.org/regular-expressions'}


class WrapsShim(object):

    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.assignments = {
            '__module__': __name__,
            '__name__': repr(wrapped),
        }

    def __getattr__(self, attr):
        return getattr(self.wrapped, attr, self.assignments[attr])


def create_html_parser(headers):

    charset = headers.get_content_charset()
    try:
        if charset and codecs.lookup(charset).name == 'iso8859-1':
            charset = 'windows-1252'
    except LookupError:
        pass

    # if charset is not specified in the Content-Type, this will be
    # None ; encoding=None produces default (ISO 8859-1) behavior.
    return HTMLParser(encoding=charset)


@composable
@cached
def parse(src):
    """ Returns an element tree create by `LXML <http://lxml.de/>`_. 
       :param src: A readable object such as a :class:`wex.response.Response`.
    """

    if not hasattr(src, 'read'):
        return src

    parser = create_html_parser(src.headers)
    etree = _ElementTree()
    try:
        # Sometimes we get URLs containing characters that aren't
        # acceptable to lxml (e.g. "http:/foo.com/bar?this=array[]").
        # When this happens lxml will quote the whole URL.
        # We don't want to have to check for this so we just always
        # quote it here and then unquote it in the `base_url` function.
        quoted_base_url = quote(src.url) if src.url else src.url
        fp = replace_invalid_ncr(src)
        etree.parse(fp, parser=parser, base_url=quoted_base_url)
    except IOError as exc:
        logger = logging.getLogger(__name__)
        logger.warning("IOError parsing %s (%s)", src.url, exc)

    root = etree.getroot()
    if root is None:
        etree._setroot(UNPARSEABLE)

    return etree


@cached
def get_base_url(root):
    if root.base_url:
        base_url = unquote(root.base_url)
    else:
        base_url = root.base_url
    return reduce(urljoin, base_href(root)[:1], base_url)


def css(expression):
    """ Returns a :func:`composable <wex.composed.composable>` callable that
        will select elements defined by a 
        `CSS selector <http://en.wikipedia.org/wiki/Cascading_Style_Sheets#Selector>`_ 
        expression.

        :param expression: The CSS selector expression.

        The callable returned accepts a :class:`wex.response.Response`, a
        list of elements or an individual element as an argument.
    """
    return parse | map_if_iter(CSSSelector(expression))


def xpath(expression, namespaces=default_namespaces):
    """ Returns :func:`composable <wex.composed.composable>` callable that will
        select elements defined by an 
        `XPath <http://en.wikipedia.org/wiki/XPath>`_ expression.

        :param expression: The XPath expression.
        :param namespaces: The namespaces.

        The callable returned accepts a :class:`wex.response.Response`, a
        list of elements or an individual element as an argument.

        For example:

        .. code-block:: pycon

            >>> from lxml.html import fromstring
            >>> tree = fromstring('<h1>Hello</h1>')
            >>> selector = xpath('//h1')

    """
    return parse | map_if_iter(XPath(expression, namespaces=namespaces))


def attrib(name, default=None):
    getter = methodcaller('get', name, default)
    mapper = map_if_iter(getter)
    return mapper


def join_to_base_url(url_getter, same_domain=True):
    def base_url_join(elem_or_tree, *args, **kwargs):
        if hasattr(elem_or_tree, 'getroottree'):
            tree = elem_or_tree.getroottree()
        else:
            tree = elem_or_tree
        base_url = get_base_url(tree.getroot())
        url = url_getter(elem_or_tree)
        # urljoin requires 'find' so give up if we don't find it (e.g. None)
        if hasattr(url, 'find'):
            joined_url = urljoin(base_url, url.strip())
            if (same_domain and 
                not joined_url.startswith(urljoin(base_url, '/'))):
                return None
            return joined_url
        return url
    return map_if_iter(base_url_join) | filter_if_iter(None)


#: A :class:`wex.composed.ComposedFunction` that returns the absolute
#: URL from an ``href`` attribute as long as it is from the same domain
#: as the base URl of the response.
href_url = join_to_base_url(methodcaller('get', 'href'), same_domain=True)
#: A :class:`wex.composed.ComposedFunction` that returns the absolute
#: URL from an ``href`` attribute.
href_any_url = join_to_base_url(methodcaller('get', 'href'), same_domain=False)
#: A :class:`wex.composed.ComposedFunction` that returns the absolute
#: URL from an ``src`` attribute.
src_url = join_to_base_url(methodcaller('get', 'src'), same_domain=False)


@composable
def normalize_space(src):
    """ Return a whitespace normalized version of its input.

    :param src: text or iterable.

    If ``src`` is iterable then a generator will be returned.
    """
    if isinstance(src, string_types):
        return space_join(src.split())
    return (normalize_space_and_join(s) for s in src)


@composable
def normalize_space_and_join(src):
    """ Return a string of space-normalized text content. """
    if isinstance(src, string_types):
        return space_join(src.split())
    normalized = [normalize_space_and_join(s) for s in src]
    return space_join((s for s in normalized if s.strip()))


@composable
def itertext(src):
    """ Iterates text from elements.

        :param src: The element or elements to iterate.
    """
    if hasattr(src, 'itertext'):
        # using '*' to exclude comments
        return (t for t in src.itertext('*'))
    elif hasattr(src, '__iter__') and not isinstance(src, string_types):
        return (itertext(s) for s in src)
    return src


#: A :class:`wex.composed.ComposedFunction` that returns the whitespace 
#: normalized text from selected elements.
text = itertext | normalize_space

#: A :class:`wex.composed.ComposedCallable` that yields text nodes.
text_nodes = itertext | flatten

#: A :class:`wex.composed.ComposedFunction` that returns the whitespace 
#: normalized text from zero or more elements joined with a space.
join_text = itertext | normalize_space_and_join



""" An extractor is a callable that returns or yields data. For example:

.. code-block:: python

    def extract(response):
        return "something"

The ``response`` parameter here is an instance of 
:class:`wex.response.Response`.

Extractors can be combined in various ways.
"""

from __future__ import absolute_import, unicode_literals, print_function
from functools import wraps
from .value import yield_values


OMITTED = object()


class Chain(object):
    """ A chain of extractors.

    The output is the output from each extractor in sequence.

    :param extractors: an iterable of extractor callables to chain

    For example an extractor function ``extract`` defined as follows:

    .. code-block:: python

        def extract1(response):
            yield "one"

        def extract2(response):
            yield "two"

        extract = Chain(extract1, extract2)

    Would produce the following extraction output:

    .. code-block:: shell

        $ wex http://example.net/
        "one"
        "two"

    """

    @property
    def __name__(self):
        return repr(self)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.extractors)

    def __init__(self, *extractors):
        self.extractors = list(extractors)

    def __call__(self, arg0, *args, **kw):
        seek = getattr(arg0, 'seek', None)
        for extractor in self.extractors:
            if seek:
                seek(0)
            values = yield_values(extractor, arg0, *args, **kw)
            for value in values:
                yield value

    def append(self, extractor):
        self.extractors.append(extractor)
        return extractor

    def insert(self, index, extractor=None):
        def decorator(func):
            self.insert(index, func)
        if extractor is None:
            return decorator
        else:
            return decorator(extractor)



def label(*label_literals_or_callables):
    """ Returns a decorator that will label the output an extractor.

    :param literals_or_callables: An iterable of labels or callables.

    Each item in ``literals_or_callables`` may be a literal or a callable.
    Any callable will called with the same parameters as the extractor
    and whatever is returned will by used as a label.

    For example an extractor function ``extract`` defined as follows:

    .. code-block:: python

        def extract1(response):
            yield "one"


        def label2(response):
            return "label2"


        extract = label("label1", label2)(extract1)

    Would produce the following extraction output:

    .. code-block:: shell

        $ wex http://example.net/
        "label1"    "label2"    "one"

    Note that if any of the labels are
    `false <https://docs.python.org/2/library/stdtypes.html#truth-value-testing>`_ 
    then no output will be generated from that extractor.
    """

    def call(label, arg0):
        return (label(arg0) if hasattr(label, '__call__') else label)

    def label_decorator(extractor):
        @wraps(extractor)
        def labelled_extractor(arg0, *args, **kw):
            labels = [call(l, arg0) for l in label_literals_or_callables]
            if not all(labels):
                # one or more missing labels so don't yield
                return

            for value in yield_values(extractor, arg0, *args, **kw):
                yield value.label(*labels)

        return labelled_extractor

    return label_decorator


class Attributes(object):
    """ A extractor that is a collection of labelled extractors.

    Extractors can be added to the collection on construction
    using keyword arguments for the labels.  For example, an 
    extractor function ``extract`` defined as follows:

    .. code-block:: python

        extract = Attributes(
            attr1 = (lambda response: "one"),
            attr2 = (lambda response: "two"),
        )

    Would produce the extraction output something like this:

    .. code-block:: shell

        $ wex http://example.net/
        "attr1"    "one"
        "attr2"    "two"

    The ordering of the attributes in the output is arbitrary.
    """

    def __init__(self, **kw):
        self.extractors = {}
        for k, v in kw.items():
            self.add(v, k)

    @property
    def __name__(self):
        return repr(self)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.extractors.keys())

    def __len__(self):
        return len(self.extractors)

    def __call__(self, *args, **kwargs):
        for name, extractor in self.extractors.items():
            for value in yield_values(extractor, *args, **kwargs):
                yield value.label(name)

    def add(self, extractor, label=None):
        """ Add an attribute extractor.

        :param callable extractor: The extractor to be added.
        :param str label: The label for the extractor.
                          This may be ``None`` in which case the
                          extractors ``__name__`` attribute will be used.

        This method returns the extractor added.  This means it can
        also be used as a decorator. For example:

        .. code-block:: python

            attrs = Attributes()

            @attrs.add
            def attr1(response):
                return "one"
        """
        if label is None:
            label = extractor.__name__
        self.extractors[label] = extractor
        return extractor



""" Helper functions for things that are iterable """

import wex.py2compat ; assert wex.py2compat
from functools import partial
from itertools import islice as islice_
from lxml.etree import _Element
from six import next, string_types
from six.moves import map, filter
from .composed import composable


class ZeroValuesError(ValueError):
    """ Zero values were found when at least one was expected. """


class MultipleValuesError(ValueError):
    """ More than one value was found when one or none were expected. """


# we never want to iterate (or flatten) things of these types
do_not_iter = string_types + (_Element,)


@composable
def flatten(item, unless_isinstance=do_not_iter):
    """ Yield items from all sub-iterables from obj. """
    stack = []
    while True:

        if not hasattr(item, '__iter__') or isinstance(item, unless_isinstance):
            yield item
        else:
            stack.append(iter(item))

        while stack:
            try:
                item = next(stack[-1])
                break
            except StopIteration:
                stack.pop()

        if not stack:
            break


def map_if_iter(func):
    @composable
    #@wraps(func)
    def wrapper(arg0):
        if not hasattr(arg0, '__iter__') or isinstance(arg0, do_not_iter):
            return func(arg0)
        else:
            return map(func, flatten(arg0))
    return wrapper


def filter_if_iter(func):
    @composable
    def wrapper(arg0):
        if not hasattr(arg0, '__iter__') or isinstance(arg0, do_not_iter):
            return arg0
        else:
            return filter(func, arg0)
    return wrapper



@composable
def first(iterable):
    """ Returns first item from an iterable.

    :param iterable: The iterable.

    If the iterable is empty then ``None`` is returned.
    """
    if not hasattr(iterable, '__iter__'):
        # turns out it isn't iterable after all
        return iterable
    i = iter(iterable)
    try:
        v0 = next(i)
    except StopIteration:
        raise ZeroValuesError
    return v0


@composable
def one(iterable):
    """ Returns an item from an iterable of exactly one element.

    If the iterable comprises zero elements then :exc:`.ZeroValuesError` is
    raised.
    If the iterable has more than one element then :exc:`.MultipleValuesError` is
    raised.
    """
    if not hasattr(iterable, '__iter__'):
        # turns out it isn't iterable after all
        return iterable
    i = iter(iterable)
    v0 = first(i)
    try:
        next(i)
        raise MultipleValuesError()
    except StopIteration:
        pass
    return v0


@composable
def one_or_none(iterable):
    """ Returns one item or ``None`` from an iterable of length one or zero.

    If the iterable is empty then ``None`` is returned.

    If the iterable has more than one element then :exc:`.MultipleValuesError` is
    raised.
    """
    try:
        return one(iterable)
    except ZeroValuesError:
        return None


def map_partial(func, *args, **kwargs):
    """ Returns a function that maps a function over a flattened iterable. """
    partial_func = partial(func, *args, **kwargs)
    @composable
    def map_partial(iterable):
        return map(partial_func, iterable)
    return map_partial


def map_flat(func, *args, **kwargs):
    return flatten | map_partial(func, *args, **kwargs)


def islice(*islice_args):
    """ Returns a function that will perform ``itertools.islice`` on its input. """
    @composable
    def islice(iterable):
        return islice_(iterable, *islice_args)
    return islice



""" The simplest way to register :mod:`extractors <wex.extractor>` is to have 
a file named ``entry_points.txt`` in the current directory.  This file should 
look something like this:

.. code-block:: cfg

    [wex]
    .example.net = mymodule:extract_from_example_net

The ``[wex]`` section heading tells Wextracto that
the following lines register extractors.

Extractors are registered using ``name = value`` pairs.
If the name starts with ``.`` then the extractor is only applied to 
responses from 
`domain names <http://en.wikipedia.org/wiki/Domain_name>`_
that match that name.
Our example would match responses from ``www.example.net`` or ``example.net``.

If the name does not start with ``.`` it will be applied responses whatever
their domain.

You can register the same extractor against multiple domain names by 
having multiple lines with the same value but different names.

This is exactly the same format and content that you would use in the 
``entry_points`` parameter for a 
`setup function <https://pythonhosted.org/setuptools/setuptools.html#new-and-changed-setup-keywords>`_, 
if and when you want to package and your extractor functions.
"""

from __future__ import absolute_import, unicode_literals, print_function
import sys
import os
import logging
import errno
from pkg_resources import EntryPoint, iter_entry_points
from six.moves.urllib_parse import urlparse
from six import itervalues
from wex.extractor import Chain


GROUP='wex'


def get_wex_entry_points_from_cwd():
    try:
        with open(os.path.join(os.getcwd(), 'entry_points.txt')) as txt:
            entry_point_map = EntryPoint.parse_map(txt.read())
        entry_points = {str(ep): ep for ep in entry_point_map[GROUP].values()}
        if os.getcwd() not in sys.path:
            sys.path.insert(0, os.getcwd())
    except IOError as exc:
        if exc.errno != errno.ENOENT:
            raise
        entry_points = {}

    return entry_points


class ExtractorFromEntryPoints(object):
    """ An extractor combining extractors loaded from entry points. """

    def __init__(self):
        self.extractors = {}
        self.wex_entry_points_from_cwd = get_wex_entry_points_from_cwd()

    def __call__(self, arg0, *args, **kw):
        hostname = urlparse(getattr(arg0, 'url', '') or '').hostname
        if hostname not in self.extractors:
            self.extractors[hostname] = self.load_extractor(hostname)
        extractor = self.extractors[hostname]
        return extractor(arg0, *args, **kw)

    def load_extractor(self, hostname):
        extractors = []
        for ep in self.iter_wex_entry_points():
            if ep.name.startswith('.') and not domain_suffix(ep, hostname):
                continue
            append_if_load_succeeded(extractors, ep)
        return Chain(*extractors)

    def iter_wex_entry_points(self):

        for ep in iter_entry_points(GROUP):
            # we don't want to load the same entry point twice
            if str(ep) in self.wex_entry_points_from_cwd:
                continue
            yield ep

        for ep in itervalues(self.wex_entry_points_from_cwd):
            yield ep


def extractor_from_entry_points():
    return ExtractorFromEntryPoints()


def domain_suffix(entry_point, name):
    return name and ('.' + name).endswith(entry_point.name)


def append_if_load_succeeded(extractors, entry_point):
    try:
        extractors.append(entry_point.load(False))
    except Exception:
        logger = logging.getLogger(__name__)
        logger.exception("Failed to load [%s] entry point '%s'",
                         GROUP, entry_point.name)




""" Extractors for URLs from 
`/robots.txt <http://en.wikipedia.org/wiki/Robots_exclusion_standard#Sitemap>`_
and `sitemaps <http://www.sitemaps.org/protocol.html>`_.
"""


from __future__ import unicode_literals, absolute_import, print_function
import wex.py2compat ; assert wex.py2compat
from lxml.etree import iterparse
from codecs import getreader
from six.moves.urllib_parse import urljoin
from wex.extractor import Chain
from wex.http import decode
from wex.url import URL


def urls_from_robots_txt(response):
    """ Yields sitemap URLs from "/robots.txt" """

    url = URL(response.request_url or response.url or '')
    if url.parsed.path != '/robots.txt':
        return

    charset = response.headers.get_content_charset()
    lines = getreader(charset or 'ISO-8859-1')(response)

    for line in lines:

        content, _, comment = line.partition('#')
        field, _, value = content.partition(':')
        if field.strip().lower() != 'sitemap':
            continue

        # we shouldn't need to urljoin but we do just in case
        joined = URL(urljoin(response.url, value.strip()))
        # set sitemap=True in fragment to help downstream processing
        yield "url", joined.update_fragment_dict(sitemap=True)


def urls_from_urlset_or_sitemapindex(response):
    """ Yields URLs from ``<urlset>`` or ``<sitemapindex>`` elements as per 
        `sitemaps.org <http://www.sitemaps.org/protocol.html>`_.
    """

    content_subtype = response.headers.get_content_subtype()
    if 'xml' not in content_subtype.split('+'):
        return

    root = None
    for _, elem in iterparse(decode(response)):

        if root is None:
            root = elem.getroottree().getroot()
            if not (root.tag.endswith('}sitemapindex') or
                    root.tag.endswith('}urlset')):
                # root element has wrong tag - give up
                break

        if elem.tag.endswith('}loc') and elem.text is not None:
            text = elem.text.strip()
            if text:
                # according to http://www.sitemaps.org/protocol.html#locdef
                # these 
                # this 
                url = URL(urljoin(response.url, text))
                if elem.getparent().tag.endswith('}sitemap'):
                    # add a sitemap=True to fragment so that
                    # in case we need to read an Atom, RSS or text file.
                    url = url.update_fragment_dict(sitemap=True)
                yield "url", url

        if elem.getparent() is root:
            # release memory for previous elements
            while elem.getprevious() is not None:
                del root[0]

#: Extractor that combines :func:`.urls_from_robots_txt` and
#: :func:`.urls_from_urlset_or_sitemapindex`.
urls_from_sitemaps = Chain(urls_from_robots_txt,
                           urls_from_urlset_or_sitemapindex)



import io
import posixpath
from functools import wraps
from ftplib import FTP
from .http import format_status_line, format_header, CRLF
from .readable import ChainedReadable


def get(url, recipe, **kw):
    """ Recipe for an FTP get. """
    timeout = kw.get('timeout', 10.0)
    ftp = FTP(url.parsed.hostname, url.parsed.username,
              url.parsed.password, timeout=timeout)
    dirname, basename = posixpath.split(url.parsed.path)
    ftp.cwd(dirname)
    return (_readable(url, RETRReadable(ftp, basename)),)


def close_on_empty(unbound):
    """ Calls 'close' on first argument when `method` return something falsey.

    The first argument is presumed to the `self`.
    """
    @wraps(unbound)
    def wrapper(self, *args):
        buf = unbound(self, *args)
        if not buf:
            self.close()
        return buf
    return wrapper


class RETRReadable(object):
    """ Just like ftplib.FTP.retrbinary, but implements read and readline. """

    def __init__(self, ftp, basename):
        self.ftp = ftp
        self.ftp.voidcmd('TYPE I')
        self.conn = ftp.transfercmd('RETR {}'.format(basename))
        self.fp = self.conn.makefile('rb')

    @close_on_empty
    def read(self, *args):
        return self.fp.read(*args)

    @close_on_empty
    def readline(self, *args):
        return self.fp.readline(*args)

    def close(self):
        self.fp.close()
        self.conn.close()
        self.ftp.voidresp()


def _readable(url, fp, **kw):
    """ Make an object that is readable by `Response`.from_file. """

    headers = io.TextIOWrapper(io.BytesIO(), encoding='utf-8', newline='\n')
    protocol = 'FTP'
    version = '1.0'
    code = 200
    reason = 'OK'
    status_line = format_status_line(protocol, version, code, reason)
    headers.write(status_line)
    headers.write(format_header('X-wex-url', url))
    headers.write(CRLF)
    headers.seek(0)

    return ChainedReadable(headers.detach(), fp)




import os
import subprocess
import tempfile
import json
from pkg_resources import resource_filename


default_settings = {
    'loadImages': False,
    'resourceTimeout': 60000,
}


phantomjs_js = os.path.abspath(resource_filename(__name__, 'phantomjs.js'))
cmd = ['phantomjs', phantomjs_js]


def request_using_phantomjs(url, method, session=None, **kw):
    phantomjs = subprocess.Popen(cmd, stdin=subprocess.PIPE)
    fifo_path = mkfifo(phantomjs)
    settings = dict(default_settings)
    url_without_fragment = url.partition('#')[0]
    request = {
        'timeout': 60000,
        'url': url_without_fragment,
        'wex_url': url,
        'wexout': fifo_path,
        'settings': settings,
        'evaluate': [],
    }

    for evaluate in method.args.get('evaluate', []):
        filename = resource_filename(*evaluate)
        request['evaluate'].append(filename)

    request_line = json.dumps(request) + '\n'
    phantomjs.stdin.write(request_line.encode('utf-8'))
    phantomjs.stdin.flush()
    try:
        yield open(fifo_path, 'rb')
    finally:
        os.unlink(fifo_path)
    assert phantomjs.wait() == 0


def mkfifo(phantomjs):
    basename = 'wex.phantomjs.{}.{}.fifo'.format(os.getpid(), phantomjs.pid)
    fifo_path = os.path.join(tempfile.gettempdir(), basename)
    os.mkfifo(fifo_path)
    return fifo_path



import threading
from functools import wraps


class Cache(dict):

    local = threading.local()

    def __enter__(self):
        self.local.cache = self

    def __exit__(self, *exc_info):
        del self.local.cache

    @classmethod
    def get(cls):
        return getattr(cls.local, 'cache', {})


def cached(f):
    """ Decorator to cache results keyed by function arguments. """

    @wraps(f)
    def wrapper(*args):
        cache = Cache.get()
        key = (f,) + args
        try:
            return cache[key]
        except TypeError:
            return f(*args)
        except KeyError:
            result = cache[key] = f(*args)
            return result

    return wrapper



""" When maintaining extractors it can be helpful to have some sample input
and output so that regression testing can be performed when we need to change
the extractors.

Wextracto supports this by using the ``--save`` or ``--save-dir`` options to
the :mod:`wex <wex.command>` command.  This option saves both the input and
output to a local directory.

This input and output can then be used for comparison with the current
extractor output.

To check compare current output against saved output run 
`py.test <http://pytest.org/>`_ like so:

.. code-block:: shell

    $ py.test saved/


"""
from __future__ import absolute_import, unicode_literals, print_function  # pragma: no cover

#
# pytest-cov reports missing coverage when things are imported *before*
# the testing has started.  That is why this module is filled with 
# function level imports and pragmas.
#

import pytest                                            # pragma: no cover

TAB = '\t'  # pragma: no cover
LF = '\n'  # pragma: no cover
MISSING = object()  # pragma: no cover

def pytest_collect_file(parent, path):                   # pragma: no cover
    from .readable import EXT_WEXIN
    if path.check(ext=EXT_WEXIN.lstrip('.')):
        return WexinFile(path, parent)


class WexinFile(pytest.File):                        # pragma: no cover
    """ A .wexin file is a stored response. """

    def __init__(self, path, parent):
        super(WexinFile, self).__init__(path, parent)
        self.extracted_values = None
        self.value_items = {}

    def collect(self):                                   # pragma: no cover
        from .output import EXT_WEXOUT
        basename = self.fspath.purebasename + EXT_WEXOUT
        path = self.fspath.dirpath().join(basename)
        self.value_items = {}
        with path.open(encoding='UTF-8') as wexout:
            # For testing purposes we keep the values as JSON strings
            # this lets us not have to worry about hashable types, etc.
            for line in wexout:
                labels, _, value = line.rpartition('\t')
                if labels in self.value_items:
                    item = self.value_items[labels]
                else:
                    item = WexoutValues(labels, self)
                    self.value_items[labels] = item
                    yield item
                item.values.add(value.strip())
        yield WexoutLabels('labels', self)

    def get_extracted_values(self, name):
        if self.extracted_values is None:
            self.extracted_values = self.extract_values()
        return self.extracted_values.get(name, set())

    def get_extracted_labels(self):
        if self.extracted_values is None:
            self.extracted_values = self.extract_values()
        return set(self.extracted_values)

    def extract_values(self):
        from .entrypoints import extractor_from_entry_points
        from .response import Response
        values = {}
        extract = extractor_from_entry_points()
        with self.fspath.open('rb') as readable:
            for value in Response.values_from_readable(extract, readable):
                for line in value.text():
                    labels, _, value = line[:-1].rpartition(TAB)
                    values.setdefault(labels, set()).add(value)
        return values


class WexoutValues(pytest.Item):                        # pragma: no cover
    """ Test that the value set is the same. """

    def __init__(self, name, parent):            # pragma: no cover
        super(WexoutValues, self).__init__(name, parent)
        self.values = set()

    def runtest(self):
        assert self.values == self.parent.get_extracted_values(self.name)


class WexoutLabels(pytest.Item):
    """ Test that we still see the same set of labels. """

    def runtest(self):
        extracted_labels = self.parent.get_extracted_labels()
        saved_labels = set(self.parent.value_items.keys())
        assert extracted_labels == saved_labels



""" The ``wex`` command extracts data from HTTP-like responses.
These responses can come from files, directories or URLs specified on the 
command line.  The command calls any :mod:`extractors <wex.extractor>` that have 
been :mod:`registered <wex.entrypoints>` and writes any data extracted
as :mod:`output <wex.output>`.

The output and input can be saved, using the ``--save`` or ``--save-dir``
command line arguments.  This is useful for 
`regression testing <http://en.wikipedia.org/wiki/Regression_testing>`_.
existing extractor functions.
The test are run using :mod:`py.test <wex.pytestplugin>`.

For the complete list of command line arguments run:

.. code-block:: shell

    $ wex --help

"""
from __future__ import absolute_import, unicode_literals, print_function
import argparse
import logging.config
from multiprocessing import cpu_count
from pkg_resources import resource_filename
from .readable import readables_from_paths
from .processpool import do
from .output import StdOut, TeeStdOut, write_values
from .value import Value
from .entrypoints import extractor_from_entry_points


default_logging_conf = resource_filename(__name__, 'logging.conf')


argparser = argparse.ArgumentParser()

argparser.add_argument(
    'paths',
    metavar='path',
    nargs='+',
    help="file, directory or url from which to extract"
)

save_group = argparser.add_argument_group("Save extraction input and output")

save_excl_group = save_group.add_mutually_exclusive_group()

save_excl_group.add_argument(
    '-s', '--save',
    action='store_const',
    dest='save_dir',
    const='saved',
    default=False,
    help="into directory 'saved'",
)

save_excl_group.add_argument(
    '--save-dir',
    dest='save_dir',
    metavar="DIR",
    help="into directory DIR",
)

process_pool_size_group = argparser.add_argument_group('Parallel processing using multiprocessing.Pool')

process_pool_size = process_pool_size_group.add_mutually_exclusive_group()
#process_pool_size = argparser.add_mutually_exclusive_group()
#process_pool_size = argparser

process_pool_size.add_argument(
    '-P',
    dest='process_pool_size',
    action="store_const",
    const=cpu_count(),
    default=1,
    help="with default pool size (default: %s)" % cpu_count(),
)

process_pool_size.add_argument(
    '--process-pool',
    dest='process_pool_size',
    metavar='N',
    type=int,
    default=1,
    help="with a pool size of N",
)

on_exc_group = argparser.add_argument_group("When an exception occurs")
on_exc = on_exc_group.add_mutually_exclusive_group()

on_exc.add_argument(
    '-x', '--exit-on-exc',
    action="store_true",
    default=False,
    help="exit with a traceback",
)

on_exc.add_argument(
    '-d', '--debug-on-exc',
    action="store_true",
    default=False,
    help="start the debugger",
)


class WriteExtractedValues(object):

    def __init__(self, context, extract):
        self.context = context
        self.extract = extract

    def __call__(self, readable):
        return write_values(self.context, readable, self.extract)


def main():

    logging.config.fileConfig(default_logging_conf,
                              disable_existing_loggers=False)

    args = argparser.parse_args()
    extract = extractor_from_entry_points()
    if args.save_dir:
        func = WriteExtractedValues(TeeStdOut, extract)
    else:
        func = WriteExtractedValues(StdOut, extract)
    Value.exit_on_exc = args.exit_on_exc
    Value.debug_on_exc = args.debug_on_exc

    readables = readables_from_paths(args.paths, args.save_dir)
    do(func, readables, pool_size=args.process_pool_size)



import requests
from six import iteritems
from six import BytesIO
from six.moves.urllib_parse import urljoin
from .py2compat import parse_headers
from .iterable import one
from .http import timeout, readable_from_response
from .etree import create_html_parser, get_base_url
from .response import Response


class ParserReadable(object):
    """ Readable that feeds a parser as it is reads. """

    def __init__(self, readable):
        self.readable = readable
        self.lines = []
        self.code = None
        self.headers = None
        self.parser = None
        self.root = None


    @classmethod
    def from_response(cls, response, url, decode_content):
        return cls(readable_from_response(response, url, decode_content))

    @property
    def name(self):
        return getattr(self.readable, 'name')

    def read(self, size):
        buf = self.readable.read(size)
        if self.parser:
            self.parser.feed(buf)
            if len(buf) < size:
                if self.root is None:
                    self.root = self.parser.close()
                    url = self.headers.get('X-wex-request-url')
                    # this sets the .base_url
                    self.root.getroottree().docinfo.URL = url
        return buf

    def readline(self, *args):
        line = self.readable.readline(*args)
        if not self.lines:
            _, _, self.code, _ = Response.parse_status_line(line)
        self.lines.append(line)
        if not line.strip():
            self.headers = parse_headers(BytesIO(b''.join(self.lines[1:])))
            if 200 <= self.code < 300:
                self.parser = create_html_parser(self.headers)
        return line

    def close(self):
        self.readable.close()


def submit_form(url, method, session=None, **kw):

    if session is None:
        session = requests.Session()
        session.stream = True

    decode_content = kw.get('decode_content', True)

    response = session.request(
        'get',
        url,
        params=method.args.get('params', None),
        data=None,
        headers=method.args.get('headers', None),
        cookies=method.args.get('cookies', None),
        timeout=timeout,
        allow_redirects=False,
    )
    readable = ParserReadable.from_response(response, url, decode_content)
    yield readable

    redirects = session.resolve_redirects(response, response.request,
                                          stream=True, timeout=timeout)

    for response in redirects:
        readable = ParserReadable.from_response(response, url, decode_content)
        yield readable

    if readable.root is None:
        return

    form_css_selector, values = one(iteritems(method.args))
    form = one(readable.root.cssselect(form_css_selector))

    if isinstance(values, dict):
        values = values.items()

    for name, value in values:
        if name in form.inputs:
            input = form.inputs[name]
        else:
            input = one(form.cssselect(name))
        if hasattr(input, 'add'):
            input.add(value)
        else:
            input.value = value

    base_url = get_base_url(form)
    form_action_url = urljoin(base_url, form.get('action', ''))

    form_method = form.method.upper()
    if form_method in ('POST', 'PUT'):
        # this implies 'application/x-www-form-urlencoded'
        data = form.form_values()
        params = None
    else:
        data = None
        params = form.form_values()

    response = session.request(
        form_method,
        form_action_url,
        params=params,
        data=data,
        headers=method.args.get('headers', None),
        cookies=method.args.get('cookies', None),
        timeout=timeout,
        allow_redirects=False,
    )
    yield readable_from_response(response, url, decode_content)

    redirects = session.resolve_redirects(response, response.request,
                                          stream=True, timeout=timeout)
    for redirect in redirects:
        yield readable_from_response(redirect, url, decode_content)



"""
URL Labelling
^^^^^^^^^^^^^

The convention for Wextracto is that any URL that should be downloaded 
is has the left-most label ``url``.  For example::

        "url"\t"http://example.net/some/url"

Data Labelling
^^^^^^^^^^^^^^

If you are extracting multiple types of data (for example people and 
addresses) then a good labelling scheme is important.

It is a good idea to label the extracted values so that you can sort them
easily using the Unix :command:`sort` command.

An example of a labelling scheme that allows this would be::

    {type}\t{identifier}\t{attribute}\t{value}

So we might end up with output that look like this::

    "person"\t"http://example.net/person/1"\t"name"\t"Tom Bombadil"
    "person"\t"http://example.net/person/1"\t"email"\t"tom1@example.net"
    "address"\t"http://example.net/address/2"\t"city"\t"New York"
    "address"\t"http://example.net/address/2"\t"postal code"\t"10001"
    "person"\t"http://example.net/person/3"\t"name"\t"Jack Sprat"
    "person"\t"http://example.net/person/3"\t"email"\t"jack3@example.net"
    "address"\t"http://example.net/address/4"\t"city"\t"London"
    "address"\t"http://example.net/address/4"\t"postal code"\t"E14 5AB"

With output like this we can easily sort and group it.
"""

from __future__ import absolute_import, unicode_literals, print_function
import os
import errno
import sys
import logging; log = logging.getLogger(__name__)
import codecs
from multiprocessing import Lock
from .response import Response
from .readable import EXT_WEXIN

EXT_WEXOUT = '.wexout'

CHUNK_SIZE = 2**8


lock = Lock()


class StdOut(object):

    if sys.stdout.encoding is None:
        stdout = codecs.getwriter('UTF-8')(sys.stdout)
    else:
        stdout = sys.stdout

    def __init__(self, readable):
        self.readable = readable
        self.buffer = []
        self.size = 0

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.flush()
        self.close()

    def close(self):
        if hasattr(self.readable, 'close'):
            self.readable.close()

    def flush(self):
        chunk = ''.join(self.buffer)
        if chunk:
            with lock:
                self.stdout.write(chunk)
                self.stdout.flush()
        self.buffer = []
        self.size = 0

    def write(self, text):
        self.buffer.append(text)
        self.size += len(text)
        if self.size > CHUNK_SIZE:
            self.flush()



class TeeStdOut(StdOut):

    def __init__(self, readable):
        super(TeeStdOut, self).__init__(readable)
        stem, ext = os.path.splitext(getattr(readable, 'name', ''))
        if stem and ext == EXT_WEXIN:
            path = stem + EXT_WEXOUT
            self.tee = codecs.open(path, 'w', 'UTF-8')
        else:
            self.tee = None

    def close(self):
        super(TeeStdOut, self).close()
        if self.tee:
            self.tee.close()

    def write(self, chunk):
        super(TeeStdOut, self).write(chunk)
        if self.tee:
            self.tee.write(chunk)

    def flush(self):
        super(TeeStdOut, self).flush()
        if self.tee:
            self.tee.flush()


def write_values(context, readable, extract):

    ret = None
    try:

        with context(readable) as writer:
            for value in Response.values_from_readable(extract, readable):
                for line in value.text():
                    writer.write(line)

    except IOError as exc:

        if exc.errno == errno.EPIPE:
            ret = SystemExit(0)
        else:
            log.exception('reading %r', readable)
            ret = SystemExit(exc)

    except Exception as exc:

        log.exception('while extracting from %r', readable)
        ret = exc
        raise

    return ret
